# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any

from airflow.configuration import conf

from airflow.providers.nomad.constants import CONFIG_SECTION

default_image = conf.get(
    CONFIG_SECTION, "default_docker_image", fallback="novakjudit/airflow-nomad-runner:0.0.5b1"
)

default_task_template: dict[str, Any] = {
    "Job": {
        "AllAtOnce": None,
        "Constraints": None,
        "CreateIndex": None,
        "Datacenters": ["dc1"],
        "ID": "example",
        "JobModifyIndex": None,
        "Meta": None,
        "ModifyIndex": None,
        "Name": "ariflow_run",
        "Namespace": None,
        "ParameterizedJob": None,
        "ParentID": None,
        "Payload": None,
        "Periodic": None,
        "Priority": None,
        "Region": None,
        "Stable": None,
        "Status": None,
        "StatusDescription": None,
        "Stop": None,
        "SubmitTime": None,
        "TaskGroups": [
            {
                "Constraints": None,
                "Count": 1,
                "EphemeralDisk": {"Migrate": None, "SizeMB": 300, "Sticky": None},
                "Meta": None,
                "Name": "airflow-execution-taskgroup",
                "RestartPolicy": {
                    "Attempts": 0,
                    "Delay": 25000000000,
                    "Interval": 300000000000,
                    "Mode": "fail",
                },
                "Tasks": [
                    {
                        "Artifacts": None,
                        "Config": {
                            "image": default_image,
                            "entrypoint": [
                                "python",
                                "-m",
                                "airflow.sdk.execution_time.execute_workload",
                                "--json-string",
                            ],
                            "args": [],
                        },
                        "Constraints": None,
                        "DispatchPayload": None,
                        "Driver": "docker",
                        "Env": {
                            "AIRFLOW_CONFIG": "/opt/airflow/config/airflow.cfg",
                            "AIRFLOW_HOME": "/opt/airflow/",
                        },
                        "KillTimeout": None,
                        "Leader": False,
                        "LogConfig": None,
                        "Meta": None,
                        "Name": "airflow-task",
                        "Resources": {
                            "CPU": 500,
                            "DiskMB": None,
                            "MemoryMB": 256,
                        },
                        "ShutdownDelay": 1000000000000,
                        "Templates": None,
                        "User": "",
                        "Vault": None,
                        "VolumeMounts": [
                            {
                                "Destination": "/opt/airflow/config",
                                "PropagationMode": "private",
                                "ReadOnly": True,
                                "SELinuxLabel": "",
                                "Volume": "config",
                            },
                            {
                                "Destination": "/opt/airflow/dags",
                                "PropagationMode": "private",
                                "ReadOnly": True,
                                "SELinuxLabel": "",
                                "Volume": "dags",
                            },
                            {
                                "Destination": "/opt/airflow/logs",
                                "PropagationMode": "private",
                                "ReadOnly": False,
                                "SELinuxLabel": "",
                                "Volume": "logs",
                            },
                        ],
                    }
                ],
                "Update": None,
                "Volumes": {
                    "config": {
                        "AccessMode": "",
                        "AttachmentMode": "",
                        "MountOptions": None,
                        "Name": "config",
                        "PerAlloc": False,
                        "ReadOnly": True,
                        "Source": "config",
                        "Sticky": False,
                        "Type": "host",
                    },
                    "dags": {
                        "AccessMode": "",
                        "AttachmentMode": "",
                        "MountOptions": None,
                        "Name": "dags",
                        "PerAlloc": False,
                        "ReadOnly": True,
                        "Source": "dags",
                        "Sticky": False,
                        "Type": "host",
                    },
                    "logs": {
                        "AccessMode": "single-node-writer",
                        "AttachmentMode": "file-system",
                        "MountOptions": None,
                        "Name": "logs",
                        "PerAlloc": False,
                        "ReadOnly": False,
                        "Source": "airflow-logs",
                        "Sticky": False,
                        "Type": "host",
                    },
                },
            }
        ],
        "Type": "batch",
        "VaultToken": None,
        "Version": None,
    }
}



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import datetime
from functools import cached_property
from typing import Any, Callable

import nomad  # type: ignore[import-untyped]
from airflow.configuration import conf
from airflow.utils.log.logging_mixin import LoggingMixin
from nomad.api.exceptions import BaseNomadException  # type: ignore[import-untyped]
from pydantic import ValidationError
from tenacity import retry, stop_after_attempt, wait_random

from airflow.providers.nomad.constants import CONFIG_SECTION
from airflow.providers.nomad.exceptions import NomadProviderException, NomadValidationError
from airflow.providers.nomad.models import (
    JobEvalStatus,
    JobInfoStatus,
    NomadJobAllocations,
    NomadJobAllocList,
    NomadJobEvalList,
    NomadJobEvaluation,
    NomadJobModel,
    NomadJobSubmission,
    NomadJobSummary,
)
from airflow.providers.nomad.utils import dict_to_lines, validate_nomad_job, validate_nomad_job_json

RETRY_NUM = conf.getint(CONFIG_SECTION, "job_submission_retry_num", fallback=3)
RETRY_MIN = conf.getint(CONFIG_SECTION, "job_submission_retry_interval_min", fallback=1)
RETRY_MAX = conf.getint(CONFIG_SECTION, "job_submission_retry_interval_max", fallback=5)


class NomadManager(LoggingMixin):
    """A layer of abstraction and encapsulate direct Nomad interactions and Nomad job management.

    Functionalities provided are equally used on Executor and Operator side.
    """

    def __init__(self):
        self.nomad_server: str = conf.get(CONFIG_SECTION, "agent_host", fallback="0.0.0.0")
        self.nomad_server_port: int = conf.getint(
            CONFIG_SECTION, "agent_server_port", fallback=4646
        )
        self.secure: bool = conf.getboolean(CONFIG_SECTION, "agent_secure", fallback=False)
        self.cert_path: str = conf.get(CONFIG_SECTION, "agent_cert_path", fallback="")
        self.key_path: str = conf.get(CONFIG_SECTION, "agent_key_path", fallback="")
        self.namespace: str = conf.get(CONFIG_SECTION, "agent_namespace", fallback="")
        self.token: str = conf.get(CONFIG_SECTION, "agent_token", fallback="")

        self.alloc_pending_timeout: int = conf.getint(
            CONFIG_SECTION, "alloc_pending_timeout", fallback=600
        )

        self.verify: bool | str
        verify = conf.get(CONFIG_SECTION, "agent_verify", fallback="")
        if verify == "true":
            self.verify = True
        elif verify == "false":
            self.verify = False
        else:
            self.verify = verify

        self.nomad: nomad.Nomad | None = None
        self.pending_jobs: dict[str, int] = {}

    def catch_nomad_exception(exc_retval: Any | None = None):  # type: ignore [reportGeneralTypeIssues]
        def decorator(f: Callable):
            def wrapped(self, *args, **kwargs):
                try:
                    return f(self, *args, **kwargs)
                except BaseNomadException as err:
                    self.log.info("Nomad error occurred: {%s}", err)
                return exc_retval

            return wrapped

        return decorator

    def ensure_nomad_client(f: Callable):  # type: ignore [reportGeneralTypeIssues, misc]
        def wrapped(self, *args, **kwargs):
            try:
                if not self.nomad:
                    self.initialize()

                return f(self, *args, **kwargs)
            except BaseNomadException as err:
                self.log.info("Nomad error occurred: {%s}", err)

        return wrapped

    @catch_nomad_exception()
    def initialize(self):
        self.nomad = nomad.Nomad(
            host=self.nomad_server,
            secure=self.secure,
            cert=(self.cert_path, self.key_path),
            verify=self.verify,  # type: ignore[reportArtumentType]
            namespace=self.namespace,
            token=self.token,
        )
        if not self.nomad:
            raise NomadProviderException("Can't initialize nomad client")

    @cached_property
    def nomad_url(self):
        protocol = "http" if not self.secure else "https"
        return f"{protocol}://{self.nomad_server}:{self.nomad_server_port}"

    @catch_nomad_exception()
    @ensure_nomad_client
    def get_nomad_job_submission(self, job_id: str) -> NomadJobSubmission | None:  # type: ignore[return]
        if not (job_status := self.nomad.job.get_job(job_id)):  # type: ignore[reportOptionalMemberAccess, union-attr]
            return  # type: ignore [return-value]
        try:
            return NomadJobSubmission.model_validate(job_status)
        except ValidationError as err:
            self.log.debug("Couldn't parse Nomad job submission info: %s %s", err, err.errors())

    @catch_nomad_exception()
    @ensure_nomad_client
    def get_nomad_job_evaluations(self, job_id: str) -> NomadJobEvalList | None:  # type: ignore[return]
        if not (job_eval := self.nomad.job.get_evaluations(job_id)):  # type: ignore[reportOptionalMemberAccess, union-attr]
            return  # type: ignore [return-value]
        try:
            return NomadJobEvaluation.validate_python(job_eval)
        except ValidationError as err:
            self.log.debug("Couldn't parse Nomad job validation output: %s %s", err, err.errors())

    @catch_nomad_exception()
    @ensure_nomad_client
    def get_nomad_job_allocation(self, job_id: str) -> NomadJobAllocList | None:  # type: ignore[return]
        if not (job_allocations := self.nomad.job.get_allocations(job_id)):  # type: ignore[reportOptionalMemberAccess, union-attr]
            return  # type: ignore [return-value]
        try:
            return NomadJobAllocations.validate_python(job_allocations)
        except ValidationError as err:
            self.log.debug("Couldn't parse Nomad job allocations info: %s %s", err, err.errors())

    @catch_nomad_exception()
    @ensure_nomad_client
    def get_nomad_job_summary(self, job_id: str) -> NomadJobSummary | None:  # type: ignore[return]
        if not (job_summary := self.nomad.job.get_summary(job_id)):  # type: ignore[reportOptionalMemberAccess, union-attr]
            return  # type: ignore [return-value]
        try:
            return NomadJobSummary.model_validate(job_summary)
        except ValidationError as err:
            self.log.debug("Couldn't parse Nomad job summary: %s %s", err, err.errors())

    @catch_nomad_exception(exc_retval="")
    @ensure_nomad_client
    def get_job_stdout(self, allocation_id: str, job_task_id: str) -> str:
        return self.nomad.client.cat.read_file(  # type: ignore[reportOptionalMemberAccess, union-attr]
            allocation_id, path=f"alloc/logs/{job_task_id}.stdout.0"
        )

    @catch_nomad_exception(exc_retval="")
    @ensure_nomad_client
    def get_job_stderr(self, allocation_id: str, job_task_id: str) -> str:
        return self.nomad.client.cat.read_file(  # type: ignore[reportOptionalMemberAccess, union-attr]
            allocation_id, path=f"alloc/logs/{job_task_id}.stderr.0"
        )

    @catch_nomad_exception(exc_retval="")
    @ensure_nomad_client
    def get_job_file(self, allocation_id: str, file_path: str) -> str:
        return self.nomad.client.cat.read_file(  # type: ignore[reportOptionalMemberAccess, union-attr]
            allocation_id, path=f"alloc/{file_path}"
        )

    @retry(
        reraise=True,
        stop=stop_after_attempt(RETRY_NUM),
        wait=wait_random(min=RETRY_MIN, max=RETRY_MAX),
    )
    def _retry_register_job(self, job_id: str, template: dict[str, Any]) -> None:
        try:
            self.nomad.job.register_job(job_id, template)  # type: ignore[reportOptionalMemberAccess, union-attr]
        except BaseNomadException as err:
            self.log.error("Couldn't run task %s (%s)", job_id, err)
            # Blind attempt to deregister, in case
            try:
                self.nomad.job.deregister_job(job_id)  # type: ignore[reportOptionalMemberAccess, union-attr]
            except Exception:
                pass
            raise err

    @ensure_nomad_client
    def register_job(self, job_model: NomadJobModel) -> str | None:
        if self.get_nomad_job_summary(job_model.Job.ID):
            raise NomadProviderException(
                f"Job {job_model.Job.ID} already exists. Job submission requires a unique ID."
            )

        template = job_model.model_dump()
        try:
            self._retry_register_job(job_model.Job.ID, template)
        except Exception as err:
            return str(err)
        self.log.info("Nomad job '%s' was submitted)", job_model.Job.ID)
        self.log.debug("Job template for '%s':\n%s)", job_model.Job.ID, str(template))
        return None

    @retry(
        reraise=True,
        stop=stop_after_attempt(RETRY_NUM),
        wait=wait_random(min=RETRY_MIN, max=RETRY_MAX),
    )
    def _retry_deregister_job(self, job_id: str) -> None:
        try:
            self.nomad.job.deregister_job(job_id)  # type: ignore[reportOptionalMemberAccess, union-attr]
        except Exception as err:
            self.log.error("Couldn't deregister job %s (%s)", job_id, err)
            raise err

    @ensure_nomad_client
    def deregister_job(self, job_id: str) -> str | None:
        try:
            self._retry_deregister_job(job_id)
        except Exception as err:
            return str(err)

        self.log.info("Nomad job '%s' was removed)", job_id)
        return None

    def timeout_expired(self, job_id: str) -> bool:
        now = int(datetime.now().timestamp())
        if not self.pending_jobs.get(job_id):
            self.pending_jobs[job_id] = now
            return False
        else:
            return now - self.pending_jobs[job_id] > self.alloc_pending_timeout

    @ensure_nomad_client
    def remove_job_if_hanging(
        self,
        job_id: str,
        job_status: NomadJobSubmission | None = None,
        job_alloc: NomadJobAllocList | None = None,
        job_eval: NomadJobEvalList | None = None,
        job_summary: NomadJobSummary | None = None,
        ignore_dead: bool = False,
    ) -> tuple[bool, str] | None:
        """Whether the job failed on Nomad side

        Typically on allocaton errors there is not feedback to Airflow, as the
        job remains in 'pending' state on Nomad side. Such issues have to be detected
        and the job execution is to be reported as failed.

        NOTE: The executor failing a job run is considered as an ERROR by Airflow.
        Despite the log message, this is the efficient way for this case. Potential Airflow-level
        task re-tries are applied corectly.

        :param key: reference to the task instance in question
        :return: either a tuple of: (True/False, additional info) or None if no data could be
                retrieved for the job
        """

        if not job_status and not (job_status := self.get_nomad_job_submission(job_id)):
            return None

        # Allocation failures: job is stuck in a 'pending' state
        if job_status.Status == JobEvalStatus.pending:
            if not self.timeout_expired(job_id):
                return False, ""

            if not job_eval and not (job_eval := self.get_nomad_job_evaluations(job_id)):
                return None

            failed_item = None
            for item in job_eval:
                if item.Status in JobEvalStatus.done_states() and item.FailedTGAllocs:
                    failed_item = item

            taskgroup_name = job_status.TaskGroups[0].Name
            if failed_item and (failed_alloc := failed_item.FailedTGAllocs.get(taskgroup_name)):  # type: ignore[reportOperationalMemberAccess, union-attr]
                self.log.info("Job %s was pending beyond timeout, stopping it.", job_id)
                self.deregister_job(job_id)  # type: ignore[reportOptionalMemberAccess, union-attr]
                error = failed_alloc.errors()
                return True, str(error)
        else:
            self.pending_jobs.pop(job_id, None)

        # Failures during job setup: job is stuck in a 'dead' state
        if not ignore_dead and job_status.Status == JobInfoStatus.dead:
            if not job_alloc:
                job_alloc = self.get_nomad_job_allocation(job_id)
            if not job_summary:
                job_summary = self.get_nomad_job_summary(job_id)

            if job_summary and job_summary.all_failed():
                self.log.info("Task %s seems dead, stopping it.", job_id)
                self.deregister_job(job_id)  # type: ignore[reportOptionalMemberAccess, union-attr]
                if job_alloc:
                    return True, str({str(alloc.errors()) for alloc in job_alloc})
                return True, ""

        return False, ""

    def parse_template_json(self, template_content: str) -> NomadJobModel | None:  # type: ignore [return]
        try:
            return validate_nomad_job_json(template_content)
        except NomadValidationError as err:
            self.log.debug("Couldn't parse template as json (%s)", err)

    @catch_nomad_exception()
    @ensure_nomad_client
    def parse_template_hcl(self, template_content: str) -> NomadJobModel | None:  # type: ignore [return]
        try:
            body = self.nomad.jobs.parse(template_content)  # type: ignore[optionalMemberAccess, union-attr]
            return validate_nomad_job({"Job": body})
        except BaseNomadException as err:
            self.log.debug("Couldn't parse template as HCL (%s)", err)

    def parse_template_content(self, template_content: str) -> NomadJobModel | None:  # type: ignore [return]
        if not template_content:
            return None

        try:
            if isinstance(template_content, dict):
                return validate_nomad_job(template_content)

            if template := self.parse_template_json(template_content):
                return template

            return self.parse_template_hcl(template_content)
        except NomadValidationError as err:
            self.log.error("Couldn't parse template '%s', (%s)", template_content, err)

    def job_all_info_str(
        self,
        job_id,
        job_summary: NomadJobSummary | None = None,
        job_alloc: NomadJobAllocList | None = None,
        job_eval: NomadJobEvalList | None = None,
    ) -> list[str]:
        output = []
        if job_summary or (job_summary := self.get_nomad_job_summary(job_id)):
            output.append("Job summary:")
            output += dict_to_lines(NomadJobSummary.model_dump(job_summary))
        if job_alloc or (job_alloc := self.get_nomad_job_allocation(job_id)):
            output.append("Job allocations info:")
            output += dict_to_lines(NomadJobAllocations.dump_python(job_alloc))
        if not job_alloc and not job_summary:
            if job_submission_info := self.get_nomad_job_submission(job_id):
                output.append("Job submission info:")
                output += dict_to_lines(NomadJobSubmission.model_dump(job_submission_info))
        if job_eval or (job_eval := self.get_nomad_job_evaluations(job_id)):
            output.append("Job evaluations:")
            output += dict_to_lines(NomadJobEvaluation.dump_python(job_eval))
        return output



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-nomad",
        "name": "Nomad",
        "description": "`Nomad <https://developer.hashicorp.com/nomad/>`__\n",
        "integrations": [
            {
                "integration-name": "Nomad",
                "external-doc-url": "https://developer.hashicorp.com/nomad/",
                "how-to-guide": ["/docs/apache-airflow-providers-nomad/operators.rst"],
                "logo": "/docs/integration-logos/Nomad.png",
                "tags": ["software"],
            },
        ],
        "executors": ["airflow.providers.nomad.executors.nomad_executor.NomadExecutor"],
        "operators": [
            {
                "integration-name": "Nomad",
                "python-modules": [
                    "airflow.providers.nomad.operators.job",
                    "airflow.providers.nomad.operators.task",
                ],
            }
        ],
        # "sensors": [
        #     {
        #         "integration-name": "Nomad",
        #         "python-modules": ["airflow.providers.nomad.sensors.nomad"],
        #     }
        # ],
        # "hooks": [
        #     {
        #         "integration-name": "Nomad",
        #         "python-modules": ["airflow.providers.nomad.hooks.nomad"],
        #     }
        # ],
        # "triggers": [
        #     {
        #         "integration-name": "Nomad",
        #         "python-modules": [
        #             "airflow.providers.nomad.triggers.pod",
        #             "airflow.providers.nomad.triggers.job",
        #         ],
        #     }
        # ],
        # "task-decorators": [
        #     {
        #         "class-name": "airflow.providers.nomad.decorators.nomad.nomad_task",
        #         "name": "nomad",
        #     },
        # ],
        "config": {
            "nomad_provider": {
                "description": None,
                "options": {
                    "parallelism": {
                        "description": "Generic Airflow executor parallelism (should be higher than 0)",
                        "version_added": "0.0.1",
                        "type": "integer",
                        "example": "128",
                        "default": "128",
                    },
                    "agent_host": {
                        "description": "Nomad server (FQDN or IP)",
                        "version_added": "0.0.1",
                        "type": "string",
                        "example": "192.168.122.226",
                        "default": "0.0.0.0",
                    },
                    "agent_secure": {
                        "description": "Whether TLS certificates are to be considered",
                        "version_added": "0.0.1",
                        "type": "boolean",
                        "example": None,
                        "default": "False",
                    },
                    "agent_cert_path": {
                        "description": "Absolute path to client certificate",
                        "version_added": "0.0.1",
                        "type": "string",
                        "example": "/absolute/path/to/certs/global-cli-nomad.pem",
                        "default": "",
                    },
                    "agent_key_path": {
                        "description": "Absolute path to client key",
                        "version_added": "0.0.1",
                        "type": "string",
                        "example": "/absolute/path/to/certs/global-cli-nomad-key.pem",
                        "default": "",
                    },
                    "agent_verify": {
                        "description": "Absolute paht to CA certificate or true/false",
                        "version_added": "0.0.1",
                        "type": "string",
                        "example": "/absolute/path/to/certs/nomad-agent-ca.pem",
                        "default": "",
                    },
                    "default_job_template": {
                        "description": "Specific .hcl or .json template to use for job submission, instead of in-built defaults",
                        "version_added": "0.0.1",
                        "type": "string",
                        "example": "/absolute/path/to/job_template.{json,hcl}",
                        "default": "",
                    },
                    "default_docker_image": {
                        "description": "Default Docker image for the default job template",
                        "version_added": "0.0.2",
                        "type": "string",
                        "example": "python:latest",
                        "default": "novakjudit/airflow-nomad-runner:latest",
                    },
                    "alloc_pending_timeout": {
                        "description": "Timeout in seconds before failed allocations may be considered as failed jobs",
                        "version_added": "0.0.1",
                        "type": "integer",
                        "example": "600",
                        "default": "600",
                    },
                    "job_submission_retry_num": {
                        "description": "Retry number for Nomad job submission and deregister",
                        "version_added": "0.0.4",
                        "type": "integer",
                        "example": "3",
                        "default": "3",
                    },
                    "job_submission_retry_interval_min": {
                        "description": "Minimum retry delay for Nomad job submission and deregister",
                        "version_added": "0.0.4",
                        "type": "integer",
                        "example": "1",
                        "default": "1",
                    },
                    "job_submission_retry_interval_max": {
                        "description": "Maximum retry delay for Nomad job submission and deregister",
                        "version_added": "0.0.4",
                        "type": "integer",
                        "example": "5",
                        "default": "5",
                    },
                    "operator_poll_delay": {
                        "description": "Time delay for Nomad Opeators supervision cycle, to check on child Nomad job",
                        "version_added": "0.0.2",
                        "type": "integer",
                        "example": "5",
                        "default": "10",
                    },
                },
            },
        },
    }



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from collections.abc import Collection
from random import randint
from typing import Any

from airflow.sdk import Context
from airflow.sdk.types import RuntimeTaskInstanceProtocol
from pydantic import ValidationError  # type: ignore[import-untyped]

from airflow.providers.nomad.exceptions import NomadTaskOperatorError
from airflow.providers.nomad.generic_interfaces.nomad_operator_interface import NomadOperator
from airflow.providers.nomad.models import (
    NomadEphemeralDisk,
    NomadJobModel,
    TaskConfig,
    Resource,
    NomadVolumeMounts,
    NomadVolumes,
)
from airflow.providers.nomad.templates.job_template import default_task_template
from airflow.providers.nomad.utils import job_id_from_taskinstance


class NomadTaskOperator(NomadOperator):
    """Nomad Operator allowing for lightweight job submission"""

    template_fields: Collection[str] = [
        "template_path",
        "template_content",
        "env",
        "image",
        "entrypoint",
        "args",
        "command",
        "task_resources",
        "ephemeral_disk",
        "volumes",
        "volume_mounts",
    ]

    def __init__(
        self,
        template_path: str | None = None,
        template_content: str | None = None,
        env: dict[str, str] | None = None,
        image: str | None = None,
        entrypoint: list[str] | None = None,
        args: list[str] | None = None,
        command: str | None = None,
        task_resources: dict[str, Any] | None = None,
        ephemeral_disk: dict[str, str] | None = None,
        volumes: dict[str, dict[str, int | bool | str]] | None = None,
        volume_mounts: list[dict[str, int | bool | str]] | None = None,
        **kwargs,
    ):
        if template_path and template_content:
            raise ValueError("Only one of 'template_content' and 'template_path' can be specified")
        self.template_content = template_content
        self.template_path = template_path
        self.image = image
        self.entrypoint = entrypoint
        self.args = args
        self.command = command
        self.env = env
        self.task_resources = task_resources
        self.ephemeral_disk = ephemeral_disk
        self.volumes = volumes
        self.volume_mounts = volume_mounts
        super().__init__(observe=True, **kwargs)

    def job_id(self, ti: RuntimeTaskInstanceProtocol):
        id_base = job_id_from_taskinstance(ti)
        rnd = randint(0, 10000)
        while self.nomad_mgr.get_nomad_job_submission(f"{id_base}-{rnd}"):
            rnd = randint(0, 10000)
        return f"{id_base}-{rnd}"

    def param_defined(self, param: str, context: Context) -> Any | None:
        if value := getattr(self, param, None):
            return value
        return context.get("params", {}).get(param)

    def prepare_job_template(self, context: Context):
        content = None
        if self.template_path:
            filepath = self.figure_path(self.template_path)
            try:
                with open(filepath) as f:
                    content = f.read()
            except (OSError, IOError) as err:
                self.log.error(f"Can't load job template ({err})")
                return
        try:
            if (
                content
                or (content := self.template_content)
                or (content := context.get("params", {}).get("template_content", ""))
            ):
                template = self.nomad_mgr.parse_template_content(content)
            else:
                template = NomadJobModel.model_validate(default_task_template)
                # Removing (Airflow SDK execution) entrypoint from default template
                config_dict = (
                    template.Job.TaskGroups[0].Tasks[0].Config.model_dump(exclude_unset=True)
                )
                config_dict.pop("entrypoint")
                template.Job.TaskGroups[0].Tasks[0].Config = TaskConfig.model_validate(config_dict)

            if not template:
                raise NomadTaskOperatorError("Nothing to execute")

            if len(template.Job.TaskGroups) > 1:
                raise NomadTaskOperatorError("NomadTaskOperator only allows for a single taskgroup")

            if len(template.Job.TaskGroups[0].Tasks) > 1:
                raise NomadTaskOperatorError("NomadTaskOperator only allows for a single task")

            if template.Job.TaskGroups[0].Count and template.Job.TaskGroups[0].Count > 1:
                raise NomadTaskOperatorError("Only a single execution is allowed (count=1)")

            if (image := self.image) or (image := context.get("params", {}).get("image")):
                template.Job.TaskGroups[0].Tasks[0].Config.image = image

            if entrypoint := self.param_defined("entrypoint", context):
                template.Job.TaskGroups[0].Tasks[0].Config.entrypoint = entrypoint

            if (args := self.args) or (args := context.get("params", {}).get("args")):
                template.Job.TaskGroups[0].Tasks[0].Config.args = args

            if command := self.param_defined("command", context):
                template.Job.TaskGroups[0].Tasks[0].Config.command = command

            if env := self.param_defined("env", context):
                template.Job.TaskGroups[0].Tasks[0].Env = env

            if resources := self.param_defined("task_resources", context):
                res_model = Resource.model_validate(resources)
                template.Job.TaskGroups[0].Tasks[0].Resources = res_model

            if volumes := self.param_defined("volumes", context):
                volumes = NomadVolumes.validate_python(volumes)
                template.Job.TaskGroups[0].Volumes = volumes

            if volume_mounts := self.param_defined("volume_mounts", context):
                volume_mounts = NomadVolumeMounts.validate_python(volume_mounts)
                template.Job.TaskGroups[0].Tasks[0].VolumeMounts = volume_mounts

            if (
                template.Job.TaskGroups[0].Tasks[0].VolumeMounts
                and template.Job.TaskGroups[0].Volumes
                and not all(
                    vol.from_volumes(template.Job.TaskGroups[0].Volumes)
                    for vol in template.Job.TaskGroups[0].Tasks[0].VolumeMounts
                )
            ):
                self.log.error(
                    "Inconsistent volume mounts: \nVolumes{%s}\nVolume Groups: {%s}",
                    template.Job.TaskGroups[0].Volumes,
                    template.Job.TaskGroups[0].Tasks[0].VolumeMounts,
                )
                return

            if ephemeral_disk := self.param_defined("ephemeral_disk", context):
                ephemeral_disk = NomadEphemeralDisk.model_validate(ephemeral_disk)
                template.Job.TaskGroups[0].EphemeralDisk = ephemeral_disk

        except ValidationError as err:
            raise NomadTaskOperatorError(f"Template validation failed: {err.errors()}")

        if not (ti := context.get("ti")):
            raise NomadTaskOperatorError(f"No task instance found in context {context}")

        template.Job.ID = self.job_id(ti)
        self.template = template



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Nomad Executor.
"""

from __future__ import annotations

import argparse
import copy
import logging
from pathlib import Path
from typing import Any

from airflow.configuration import conf
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.utils.log.file_task_handler import FileTaskHandler
from airflow.utils.log.logging_mixin import remove_escape_codes
from airflow.utils.state import TaskInstanceState
from pydantic import ValidationError

from airflow.providers.nomad.constants import CONFIG_SECTION
from airflow.providers.nomad.exceptions import NomadProviderException, NomadValidationError
from airflow.providers.nomad.generic_interfaces.executor_interface import ExecutorInterface
from airflow.providers.nomad.log import NomadLogHandler
from airflow.providers.nomad.manager import NomadManager
from airflow.providers.nomad.models import NomadJobModel
from airflow.providers.nomad.templates.job_template import default_task_template
from airflow.providers.nomad.utils import (
    job_id_from_taskinstance_key,
    job_task_id_from_taskinstance_key,
)

logger = logging.getLogger(__name__)


class NomadExecutor(ExecutorInterface):
    """Executor for Nomad."""

    RUNNING_POD_LOG_LINES = 100
    supports_ad_hoc_ti_run: bool = True

    def __init__(self):
        self.parallelism: int = conf.getint(CONFIG_SECTION, "parallelism", fallback=128)
        self.nomad_mgr = NomadManager()
        super().__init__(parallelism=self.parallelism)

    def start(self) -> None:
        """Start the executor."""
        self.log.info("Starting Nomad executor")
        self.nomad_mgr.initialize()

    def _get_job_template(self) -> NomadJobModel | None:
        if not (job_tpl_loc := conf.get(CONFIG_SECTION, "default_job_template", fallback="")):
            return None

        job_tpl_path = Path(job_tpl_loc)
        if not job_tpl_path.is_file():
            self.log.error("Configured template %s is not a file", job_tpl_path)
            return None

        job_template = None

        try:
            with open(job_tpl_path) as file:
                content = file.read()
        except (IOError, OSError) as err:
            self.log.error("Couldn't open job template file %s (%s)", job_tpl_path, err)
            return  # type: ignore [return-value]

        try:
            if job_tpl_path.suffix == ".json":
                job_template = self.nomad_mgr.parse_template_json(content)
            elif job_tpl_path.suffix == ".hcl":
                job_template = self.nomad_mgr.parse_template_hcl(content)
        except NomadValidationError as err:
            self.log.error("Couldn't parse job template %s (%s)", job_tpl_path, err)

        return job_template

    def prepare_job_template(self, key: TaskInstanceKey, command: list[str]) -> dict[str, Any]:
        """Adjutst template to suit upcoming job execution

        :param key: reference to the task instance in question
        :return: job template as as dictionary
        """
        job_id = job_id_from_taskinstance_key(key)
        job_task_id = job_task_id_from_taskinstance_key(key)

        if job_model := self._get_job_template():
            job_model.Job.TaskGroups[0].Tasks[0].Config.args = [
                "python",
                "-m",
                "airflow.sdk.execution_time.execute_workload",
                "--json-string",
            ] + command

        if not job_model:
            job_template = copy.deepcopy(default_task_template)
            job_model = NomadJobModel.model_validate(job_template)
            job_model.Job.TaskGroups[0].Tasks[0].Config.args = command
            job_model.Job.Name = f"airflow-run-{job_task_id}-{key[3]}"

        if not job_model:
            raise NomadProviderException("Couldn't retrieve job template")

        job_model.Job.TaskGroups[0].Tasks[0].Name = job_task_id
        job_model.Job.ID = job_id

        self.log.debug(
            f"Command running: python -m airflow.sdk.execution_time.execute_workload --json-string '{command[0]}'\n"
        )
        return job_model.model_dump(exclude_unset=True)

    def run_job(self, job_template: dict[str, Any] | None) -> str | None:
        """Execute the job defined by a potential job template

        :param: Job template corresponding to the job
        :return: No news is good news, or the error that occured on execution attempt
        """
        if not job_template:
            raise NomadProviderException("Job template missing")

        try:
            job_model = NomadJobModel.model_validate(job_template)
        except ValidationError:
            raise NomadProviderException("Job template doesn't comply to expected format ({err})")

        return self.nomad_mgr.register_job(job_model)

    def remove_job_if_hanging(self, key: TaskInstanceKey) -> tuple[TaskInstanceState, str] | None:
        """Whether the job failed outside of the Airflow context

        :param key: reference to the task instance in question
        :return: either a tuple of: (task status to set (typically: FAILED), additional info)
                 or None if no data could be retrieved for the job
        """
        job_id = job_id_from_taskinstance_key(key)
        if not (outcome := self.nomad_mgr.remove_job_if_hanging(job_id)):
            return None

        killed, error_msg = outcome
        if killed:
            return TaskInstanceState.FAILED, error_msg
        return None

    def _get_task_log(
        self, ti: TaskInstance, try_number: int, stderr=False
    ) -> tuple[list[str], list[str]]:
        messages = []
        log = []
        logtype = "error" if stderr else "standard"
        try:
            messages.append(
                f"Attempting to fetch {logtype} logs for task {ti.key} through Nomad API (attempts: {try_number})"
            )
            messages_received, logs_received = self.retrieve_logs(ti.key, stderr=stderr)
            messages += messages_received

            for line in logs_received:
                log.append(remove_escape_codes(line))
            if log:
                messages.append(f"Found {logtype} logs for running job via Nomad API")
        except Exception as e:
            messages.append(f"Reading {logtype} logs failed: {e}")

        # In case the task didn't even make it to be submitted, we may be able to get info about reasons from Nomad
        if not log and not stderr:
            job_id = job_id_from_taskinstance_key(ti.key)
            if job_info := self.nomad_mgr.job_all_info_str(job_id):
                messages.append("Nomad job evaluations retrieved")
                log.append(
                    "No task logs found, but the following information was retrieved from Nomad:"
                )
                log += job_info

        return messages, log

    def get_task_log(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        """Universal way to retreive logs

        Retrieving logs so that it's compatible both with FileTaskHandler, and
        leaves space for NomadLogHandler grouping feature.

        :params ti: TaskInstance in question
        :return: a list of operational messages together with the logs (as a list of entries)
        """
        messages, logs = self._get_task_log(ti, try_number)
        if (
            conf.get("logging", "task_log_reader", fallback=FileTaskHandler.name)
            != NomadLogHandler.name
        ):
            messages_err, logs_err = self._get_task_log(ti, try_number, stderr=True)
            if logs_err:
                logs = logs + logs_err
                messages = messages + messages_err
        return messages, logs

    def get_task_stderr(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        return self._get_task_log(ti, try_number, stderr=True)

    def retrieve_logs(self, key: TaskInstanceKey, stderr=False) -> tuple[list[str], list[str]]:
        # Note: this method is executed by the FileTaskHandler and not the Scheduler
        # We have no access to the running NomadExecutor's state
        if not self.nomad_mgr.nomad:
            self.start()

        messages = []
        logs = ""
        job_id = job_id_from_taskinstance_key(key)
        allocations = self.nomad_mgr.get_nomad_job_allocation(job_id)

        if not isinstance(allocations, list):
            messages.append("Unexpected result from Nomad API allocations query")
        elif len(allocations) == 0:
            messages.append(f"No allocations found for {job_id}")
        else:
            for allocation in allocations:
                if len(allocations) > 1:
                    logs += f"\nAllocation ID {allocation.ID}:\n"

                # Normally this is single element, as the executor insists on a single task
                for task_name in allocation.TaskStates:
                    if len(allocation.TaskStates) > 1:
                        logs += f"\nTask Name {task_name}:\n"

                    if stderr:
                        logs += self.nomad_mgr.get_job_stderr(allocation.ID, task_name)
                    else:
                        logs += self.nomad_mgr.get_job_stdout(allocation.ID, task_name)

        return messages, logs.splitlines()  # type: ignore[reportReturnType]


def _get_parser() -> argparse.ArgumentParser:  # pragma: no-cover
    """
    Generate documentation; used by Sphinx.

    :meta private:
    """
    return NomadExecutor._get_parser()



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Logging module to fetch logs via the Nomad API"""

import logging
from collections.abc import Callable
from itertools import chain
from types import GeneratorType
from typing import Generator, cast

from airflow.executors.base_executor import BaseExecutor
from airflow.executors.executor_loader import ExecutorLoader
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancehistory import TaskInstanceHistory
from airflow.utils.log.file_task_handler import (
    LegacyProvidersLogType,
    LogHandlerOutputStream,
    LogMetadata,
    LogSourceInfo,
    StructuredLogMessage,
    StructuredLogStream,
)
from airflow.utils.state import TaskInstanceState

logger = logging.getLogger(__name__)


class ExecutorLogLinesHandler(logging.Handler):
    """Log handler retrieve logs directly from Nomad

    Log format: Airflow's StructuredLogMessage stream
    """

    name = "executor_log_lines_handler"
    executor_instances: dict[str, BaseExecutor] = {}
    DEFAULT_EXECUTOR_KEY = "_default_executor"

    def __init__(self):
        super().__init__()
        self.handler: logging.Handler | None = None

    def emit(self, record):
        if self.handler:
            self.handler.emit(record)

    def flush(self):
        if self.handler:
            self.handler.flush()

    def close(self):
        if self.handler:
            self.handler.close()

    def _get_executor_function(
        self, ti: TaskInstance, function_name: str
    ) -> Callable[[TaskInstance, int], tuple[list[str], list[str]]] | None:
        executor_name = ti.executor or self.DEFAULT_EXECUTOR_KEY
        executor = self.executor_instances.get(executor_name)
        if executor is not None and hasattr(executor, function_name):
            function = getattr(executor, function_name, None)
        else:
            if executor_name == self.DEFAULT_EXECUTOR_KEY:
                self.executor_instances[executor_name] = ExecutorLoader.get_default_executor()
            else:
                self.executor_instances[executor_name] = ExecutorLoader.load_executor(executor_name)
            function = getattr(self.executor_instances[executor_name], function_name, None)

        if callable(function):
            return function  # type: ignore[reportReturnType]
        return None

    def _read(
        self,
        ti: TaskInstance | TaskInstanceHistory,
        try_number: int,
        metadata: LogMetadata | None = None,
    ) -> tuple[LogHandlerOutputStream | LegacyProvidersLogType, LogMetadata]:
        """Retieving logs from executor, parsing it to a StructuredLogMessage stream."""
        logger.info("Collecting Nomad logs")
        sources: LogSourceInfo = []
        err_sources: LogSourceInfo = []
        source_list: list[str] = []

        logs: list[str] = []
        stderr: list[str] = []

        # Stdout
        executor_get_task_log = self._get_executor_function(ti, "get_task_log")
        if not executor_get_task_log:
            logger.error("Executor doesn't support 'get_task_log()' call")
            return (iter([StructuredLogMessage(event="")]), {"end_of_log": True})

        response = executor_get_task_log(ti, try_number)
        if response:
            sources, logs = response
        if sources:
            source_list.extend(sources)

        # Stderr
        if executor_get_task_stderr := self._get_executor_function(ti, "get_task_stderr"):
            response = executor_get_task_stderr(ti, try_number)
            if response:
                err_sources, stderr = response
            if err_sources:
                source_list.extend(err_sources)

        # out_stream: LogHandlerOutputStream = _interleave_logs((y for y in logs))

        def geneate_log_stream(logs: list[str]) -> StructuredLogStream:
            for line in logs:
                if not line:
                    continue
                try:
                    yield StructuredLogMessage.model_validate_json(line)
                except:  # noqa
                    yield StructuredLogMessage(event=str(line))

        out_stream: LogHandlerOutputStream = geneate_log_stream(logs)
        err_stream: LogHandlerOutputStream = geneate_log_stream(stderr)

        # Same as for FileTaskHandler, add source details as a collapsible group
        footer = StructuredLogMessage(event="::endgroup::")
        header = [
            StructuredLogMessage(
                event="::group::Log message source details",
                sources=source_list,  # type: ignore[call-arg]
            ),
            footer,
        ]
        end_of_log = ti.try_number != try_number or ti.state not in (
            TaskInstanceState.RUNNING,
            TaskInstanceState.DEFERRED,
        )
        if stderr:
            header_log = [StructuredLogMessage(event="::group::Task logs")]
            header_err = [StructuredLogMessage(event="::group::Errors outside of task execution")]
            out_stream = chain(
                header, header_log, out_stream, [footer], header_err, err_stream, [footer]
            )
        else:
            out_stream = chain(header, out_stream)
        log_pos = len(logs)
        if metadata and "log_pos" in metadata:
            log_pos = metadata["log_pos"] + log_pos

        return out_stream, {
            "end_of_log": end_of_log,
            "log_pos": log_pos,
        }

    def read(
        self,
        task_instance: TaskInstance | TaskInstanceHistory,
        try_number: int | None = None,
        metadata: LogMetadata | None = None,
    ) -> tuple[LogHandlerOutputStream, LogMetadata]:
        """
        Read logs of given task instance from local machine.

        :param task_instance: task instance object
        :param try_number: task instance try_number to read logs from. If None
                            it returns the log of task_instance.try_number
        :param metadata: log metadata, can be used for steaming log reading and auto-tailing.
        :return: a list of listed tuples which order log string by host
        """
        if try_number is None:
            try_number = task_instance.try_number

        if try_number == 0 and task_instance.state == TaskInstanceState.SKIPPED:
            logs = [
                StructuredLogMessage(  # type: ignore[call-arg]
                    event="Task was skipped, no logs available."
                )
            ]
            return chain(logs), {"end_of_log": True}

        if try_number is None or try_number < 1:
            logs = [
                StructuredLogMessage(  # type: ignore[call-arg]
                    level="error",  # type: ignore[reportCallIssue]
                    event=f"Error fetching the logs. Try number {try_number} is invalid.",
                )
            ]
            return chain(logs), {"end_of_log": True}

        # compatibility for es_task_handler and os_task_handler
        read_result = self._read(task_instance, try_number, metadata)
        out_stream, metadata = read_result
        # If the out_stream is None or empty, return the read result
        if not out_stream:
            out_stream = cast("Generator[StructuredLogMessage, None, None]", out_stream)
            return out_stream, metadata
        if isinstance(out_stream, (chain, GeneratorType)):
            out_stream = cast("Generator[StructuredLogMessage, None, None]", out_stream)
            return out_stream, metadata
        if isinstance(out_stream, list) and isinstance(out_stream[0], StructuredLogMessage):
            out_stream = cast("list[StructuredLogMessage]", out_stream)
            return (log for log in out_stream), metadata
        else:
            raise TypeError(
                "Invalid log stream type. Expected a generator of StructuredLogMessage, list of StructuredLogMessage, list of str or str."
                f" Got {type(out_stream).__name__} instead."
            )
